#!/usr/lib/legOS25/compiler/project/perl -w

use LegoDefs;
use strl2emDefs;

# open input file and parse it
$input_file = $ARGV[$#ARGV];
if (!open(INPUT, "<$input_file")) {
	die "Can't open input file\n";
}

# open output file for support functions
# $output_file = "$input_file.out";
# if (!open(OUTPUT, ">$output_file")) {
# 	die "Can't open output file for general output\n";
# }

# open output file for support functions
$output_file1 = "$input_file.c";
if (!open(OUTSUPPORT, ">$output_file1")) {
	die "Can't open output file for support functions\n";
}

# open output file for listing commands
$output_file2 = "strl2em.temp";
if (!open(OUTLIST, ">$output_file2")) {
	die "Can't open output file for listing commands\n";
}

# write the header of the output file
print OUTSUPPORT "#include <stdlib.h>\n";
print OUTSUPPORT "#include <sys/em.h>\n";
print OUTSUPPORT "#include <conio.h>\n";
print OUTSUPPORT "#include <unistd.h>\n";
print OUTSUPPORT "#include <time.h>\n";
print OUTSUPPORT "#include <dsound.h>\n";
print OUTSUPPORT "#include <dmotor.h>\n";
print OUTSUPPORT "#include <dsensor.h>\n\n\n";

$func_num = 1;

while (<INPUT>) {

 PARSE: {

	# Remove the trailing newline character
	chomp($_);
	# Remove the leading spaces if not an empty line, if empty then go to next line
	if ($_ =~ /^\s*$/) {
		next PARSE;
	} else {
		$_ =~ s/\s*//;
	};

	# Module Name
	$_ =~ /^module \s* (\w+) \s* :/x && do {
		$ModuleName = $1;
		# print OUTPUT "$ModuleName\n";
		next PARSE;
	};

	# Inputs
	$_ =~ /^input \s*/x && do {
		my $inputs = $';
		# print STDERR "$inputs\n";
		$inputs =~ s/[,|;|\s]+/,/g;
		# print STDERR "$inputs\n";
		@input_variables = split(/,/,$inputs);
		foreach (@input_variables) {
			if (exists $ValidInputs{$_}) {
				# print OUTPUT "Input Variable = $_\n";
				$DeclaredInputs{$_} = $ValidInputs{$_};
			} else {
				die "Unknown input variable: $_\n";
			};
		};
		next PARSE;
	};
	
	# Outputs (functions and variables)
	$_ =~ /^output \s*/x && do {
		my $outputs = $';
		$outputs =~ s/[,|;]/,/g;
		$outputs =~ s/[\s]+//g;
		$outputs =~ /([,\w]+) : ([\w]+), /x && do {
			my $var_type = $2;
			@output_variables = split(/,/,$1);
			foreach (@output_variables) {
				if (exists $ValidOutputs{$_}) {
					# print OUTPUT "Output Variable = $_ $var_type \n";
					$DeclaredOutputs{$_} = $ValidOutputs{$_};
				} else {
					die "Unknown output variable: $_\n";
				};
			};
		next PARSE;
		};
		my @temp_output_functions = split(/,/,$outputs);
		foreach (@temp_output_functions) {
			$_ =~ /(\w+) \( (\w+) \) /x;
			# print OUTPUT "Output Function = $1 $2\n";
			$DeclaredOutputs{$1} = $ValidOutputs{$1};
		}
		next PARSE;
	};

	# Constants
	$_ =~ /^constant \s* /x && do {
		my $constants = $';
		$constants =~ s/[,|;]/,/g;
		$constants =~ s/[\s]+//g;
		$constants =~ /([,\w=]+) : ([\w]+), /x && do {
			my $const_type = $2;
			@constants = split(/,/,$1);
			foreach (@constants) {
				if ($_ =~ /(\w+) = (\w+)/x) {
					if (exists $ValidConstants{$1}) {
						$DeclaredConstants{$1} = $2;
						$ValidConstants{$1} = $2;
						# print OUTPUT "Constant = $1 $const_type $ValidConstants{$1}\n";
					} else {
						die "Unknown constant: $1\n";
					}
				} elsif ($_ =~ /(\w+) /x) {
					if (exists $ValidConstants{$1}) {
						$DeclaredConstants{$1} = $ValidConstants{$1};
						# print OUTPUT "Constant = $1 $const_type $ValidConstants{$1}\n";
					} else {
						die "Unknown constant: $1\n";
					};
				};
			};
		next PARSE;
		};
	};	

	# Emit Function
	$_ =~ /^emit \s* /x && do {
		my $arguments = $';
		$arguments =~ s/[\s]+//g;
		$arguments =~ /(\w+) \( (.+) \) ;/x;
		# print OUTPUT "Emit CALL $1 $2\n";
		my $output_variable = $1;
		# Write support and list function
		if (exists $DeclaredOutputs{$1}) {
			if ($2 =~ /^(\d+)$/) {
				$value = $1;
			} elsif ($2 =~ /^(\w+)$/) {
				if (exists $DeclaredConstants{$1}) {
					$value = $ValidConstants{$1};
				} else {
					die "Constant not declared: $1\n";
				}
			}
			my $lc_variable = lc $output_variable;	
			print OUTSUPPORT "void function$func_num() {\n";
			print OUTSUPPORT "	$lc_variable($value);\n";
			print OUTSUPPORT "}\n";
			print OUTSUPPORT "\n";
			print OUTLIST "$EMIT $ValidOutputs{$output_variable} &function$func_num\n";
			$func_num = $func_num + 1;
		} else {
			die "Output not declared: $1\n";
		}
		next PARSE;
	};

	# Await Function
	$_ =~ /^await \s* /x && do {
		my $arguments = $';
		$arguments =~ s/[\s]+//g;
		$arguments =~ /(\w+) ;/x;
		# print OUTPUT "Await FUTURE $1\n";
		if (exists $DeclaredInputs{$1}) {
			print OUTLIST "$AWAIT $ValidInputs{$1} 0\n";
		} else {
			die "Input not declared: $1\n";
		}
		next PARSE;
	};

	# Loop Function
	$_ =~ /^loop \s* /x && do {
		# print OUTPUT "Loop JMP\n";
		print OUTLIST "$LOOP 0 0\n";
		next PARSE;
	};

	# End Loop Function
	$_ =~ /^end \s* loop \s* /x && do {
		# print OUTPUT "End Loop JMP\n";
		print OUTLIST "$END_LOOP 0 0\n";
		next PARSE;
	};

	# [ Function (Begin block)
	$_ =~ /^\[ \s* /x && do {
		# print OUTPUT "[ Begin Block\n";
		print OUTLIST "$START_BLOCK 0 0\n";
		next PARSE;
	};

	# ] Function (End block)
	$_ =~ /^\] \s* /x && do {
		# print OUTPUT "] End Block\n";
		print OUTLIST "$END_BLOCK 0 0\n";
		next PARSE;
	};

	# || Function (Parallel operator)
	$_ =~ /^|| \s* /x && do {
		# print OUTPUT "|| Parallel Operator\n";
		print OUTLIST "$PARALLEL 0 0\n";
		next PARSE;
	};

	print "Failed: $_\n";
 }
}

close INPUT;
# close OUTPUT;
close OUTSUPPORT;
close OUTLIST;
exit(0);

